
Require Import Tactics.
Require Import Sequence.
Require Import Relation.
Require Import Subst.
Require Import Static.
Require Import Dynamic.
Require Import TSubst.
Require Import Formation.
Require Import Functionality.
Require Import Regularity.
Require Import Singleton.
Require Import Principal.
Require Import Declarative.
Require Import Congruence.


(* Target terms *)

Inductive target : (* tm *) term -> Prop :=
| target_var {i}
    : target (var i)

| target_star
    : target tm_star

| target_lam {t e}
    : target e
      -> target (tm_lam t e)

| target_app {e1 e2}
    : target e1
      -> target e2
      -> target (tm_app e1 e2)

| target_pair {e1 e2}
    : target e1
      -> target e2
      -> target (tm_pair e1 e2)

| target_pi1 {e}
    : target e
      -> target (tm_pi1 e)

| target_pi2 {e}
    : target e
      -> target (tm_pi2 e)

| target_plam {k e}
    : target e
      -> target (tm_plam k e)

| target_papp {e c}
    : target e
      -> target c
      -> target (tm_papp e c)

| target_pack {c e k t}
    : target c
      -> target e
      -> target (tm_pack c e k t)

| target_unpack {e1 e2}
    : target e1
      -> target e2
      -> target (tm_unpack e1 e2)

| target_fix {t e}
    : target e
      -> target (tm_fix t e)

| target_lett {e1 e2}
    : target e1
      -> target e2
      -> target (tm_lett e1 e2)

| target_lam_cn {k c}
    : target c
      -> target (cn_lam k c)

| target_app_cn {c1 c2}
    : target c1
      -> target c2
      -> target (cn_app c1 c2)

| target_pair_cn {c1 c2}
    : target c1
      -> target c2
      -> target (cn_pair c1 c2)

| target_pi1_cn {c}
    : target c
      -> target (cn_pi1 c)

| target_pi2_cn {c}
    : target c
      -> target (cn_pi2 c)

| target_star_cn
    : target cn_star

| target_unit
    : target cn_unit

| target_arrow {t1 t2}
    : target t1
      -> target t2
      -> target (cn_arrow t1 t2)

| target_prod {t1 t2}
    : target t1
      -> target t2
      -> target (cn_prod t1 t2)

| target_all {k t}
    : target t
      -> target (cn_all k t)

| target_exists {k t}
    : target t
      -> target (cn_exists k t)

| target_nonsense
    : target nonsense
.


Inductive targetsub : sub -> Prop :=
| targetsub_dot {m s}
    : target m
      -> targetsub s
      -> targetsub (dot m s)

| targetsub_sh {i}
    : targetsub (sh i)
.


Lemma shift_term_target :
  forall n e,
    target e
    -> target (shift_term n e).
Proof.
intros n e Htar.
revert n.
induct Htar;
try (intros;
     simp_sub;
     eauto using target;
     done); [].

(* var *)
intros i n.
unfold shift_term.
simp_sub.
unfold shift_var.
forget (Compare_dec.lt_dec i n) as b.
destruct b; apply target_var; done.
Qed.


Lemma shift_sub_target :
  forall s,
    targetsub s
    -> targetsub (shift_sub s).
Proof.
intros s H.
induct H.

(* dot *)
intros m s Hm _ IH.
simpl.
#2 apply targetsub_dot.
  apply shift_term_target; auto; done.

  apply IH; done.

(* sh *)
intros i.
simpl.
apply targetsub_sh; done.
Qed.


Lemma under1_target :
  forall s,
    targetsub s
    -> targetsub (dot (var 0) (compose s sh1)).
Proof.
intros s H.
apply targetsub_dot; auto using target_var; [].
rewrite <- shift_sub_eq_compose_sh; [].
apply shift_sub_target; auto; done.
Qed.


Lemma under2_target :
  forall s,
    targetsub s
    -> targetsub (dot (var 0) (dot (var 1) (compose s (sh 2)))).
Proof.
intros s H.
replsub (dot (var 0) (dot (var 1) (compose s (sh 2)))) with (dot (var 0) (compose (dot (var 0) (compose s sh1)) sh1)).
auto using under1_target.
Qed.

Lemma under_target :
  forall i s,
    targetsub s
    -> targetsub (under i s).
Proof.
intros i s Hs.
revert s Hs.
induct i.

(* 0 *)
intros s Hs.
simp_sub.
assumption.

(* S *)
intros i IH s Hs.
rewrite -> under_succ; [].
apply under1_target; auto; done.
Qed.


Lemma under_id_target :
  forall i, targetsub (under i id).
Proof.
intros i.
apply under_target; [].
apply targetsub_sh; done.
Qed.


Lemma subst_target :
  forall s e,
    targetsub s
    -> target e
    -> target (subst s e).
Proof.
intros s e Hs He.
revert s Hs.
induct He;
try (intros;
     simp_sub;
     eauto using target, under1_target, under2_target;
     done); [].

(* var *)
intros i s Hs.
simp_sub.
revert i.
induct Hs.
  (* dot *)
  intros m s Hm _ IH i.
  destruct i as [| i'].
    simp_sub; auto; done.

    simp_sub; auto; done.

  (* sh *)
  intros i j.
  simp_sub.
  apply target_var; done.
Qed.


Lemma compose_target :
  forall s s',
    targetsub s
    -> targetsub s'
    -> targetsub (compose s s').
Proof.
intros s1 s2 Hs1 Hs2.
revert s2 Hs2.
induct Hs1.

(* dot *)
intros m s1 Hm Hs1 IH s2 Hs2.
simp_sub.
#2 apply targetsub_dot.
  apply subst_target; auto; done.

  apply IH; auto; done.

(* sh *)
intros i s2 Hs2.
revert i.
induct Hs2.
  (* dot *)
  intros m s Hm Hs IH i.
  destruct i as [| i'].
    simp_sub.
    apply targetsub_dot; auto; done.

    simp_sub.
    apply IH; auto; done.

  (* sh *)
  intros i j.
  simp_sub.
  apply targetsub_sh; done.
Qed.


Lemma ofc_target :
  forall G c k,
    ofc G c k
    -> target c.
Proof.
intros G c k Hof.
induct Hof; try (intros; eauto using target; done); [].
intros G k1 k2 c _ Htar _ _ _.
invert Htar; auto; done.
Qed.


(* Compatibility *)

Inductive ttcc (Rk : kreln) (Rc : creln) (Rt : treln) (Rv : treln) : treln :=
| ttcc_lam {G t1 t1' t2 e e'}
    : Rc G t1 t1' kd_type
      -> Rt (G; cl_tm t1) e e' (shift1 t2)
      -> ttcc Rk Rc Rt Rv G (tm_lam t1 e) (tm_lam t1' e') (cn_arrow t1 t2)

| ttcc_app {G t1 t2 e1 e1' e2 e2'}
    : Rt G e1 e1' (cn_arrow t1 t2)
       -> Rt G e2 e2' t1
       -> ttcc Rk Rc Rt Rv G (tm_app e1 e2) (tm_app e1' e2') t2

| ttcc_pair {G t1 t2 e1 e1' e2 e2'}
    : Rt G e1 e1' t1
      -> Rt G e2 e2' t2
      -> ttcc Rk Rc Rt Rv G (tm_pair e1 e2) (tm_pair e1' e2') (cn_prod t1 t2)

| ttcc_pi1 {G t1 t2 e e'}
    : Rt G e e' (cn_prod t1 t2)
      -> ttcc Rk Rc Rt Rv G (tm_pi1 e) (tm_pi1 e') t1

| ttcc_pi2 {G t1 t2 e e'}
    : Rt G e e' (cn_prod t1 t2)
      -> ttcc Rk Rc Rt Rv G (tm_pi2 e) (tm_pi2 e') t2

| ttcc_plam {G k k' t e e'}
    : Rk G k k'
      -> Rv (G; cl_cn k) e e' t
      -> ttcc Rk Rc Rt Rv G (tm_plam k e) (tm_plam k' e') (cn_all k t)

| ttcc_papp {G k t e e' c c'}
    : Rt G e e' (cn_all k t)
      -> Rc G c c' k
      -> ttcc Rk Rc Rt Rv G (tm_papp e c) (tm_papp e' c') (subst1 c t)

| ttcc_pack {G k k' t t' c c' e e'}
    : Rc G c c' k
      -> Rk G k k'
      -> Rt G e e' (subst1 c t)
      -> Rc (G; cl_cn k) t t' kd_type
      -> ttcc Rk Rc Rt Rv G (tm_pack c e k t) (tm_pack c' e' k' t') (cn_exists k t)

| ttcc_unpack {G k t t' e1 e1' e2 e2'}
    : Rt G e1 e1' (cn_exists k t)
      -> Rt (G; cl_cn k; cl_tm t) e2 e2' (subst (sh 2) t')
      -> ofc G t' kd_type
      -> ttcc Rk Rc Rt Rv G (tm_unpack e1 e2) (tm_unpack e1' e2') t'

| ttcc_fix {G t t' e e'}
    : Rc G t t' kd_type
      -> Rt G e e' (cn_arrow (cn_arrow cn_unit t) t)
      -> ttcc Rk Rc Rt Rv G (tm_fix t e) (tm_fix t' e') t

| ttcc_lett {G t1 t2 e1 e1' e2 e2'}
    : Rt G e1 e1' t1
      -> Rt (G; cl_tm t1) e2 e2' (shift1 t2)
      -> ttcc Rk Rc Rt Rv G (tm_lett e1 e2) (tm_lett e1' e2') t2

| ttcc_equiv {G t1 t2 e e'}
    : Rt G e e' t1
      -> eqc G t1 t2 kd_type
      -> ttcc Rk Rc Rt Rv G e e' t2
.


Inductive tvcc (Rk : kreln) (Rc : creln) (Rt : treln) (Rv : treln) : treln :=
| tvcc_lam {G t1 t1' t2 e e'}
    : Rc G t1 t1' kd_type
      -> Rt (G; cl_tm t1) e e' (shift1 t2)
      -> tvcc Rk Rc Rt Rv G (tm_lam t1 e) (tm_lam t1' e') (cn_arrow t1 t2)

| tvcc_pair {G t1 t2 e1 e1' e2 e2'}
    : Rv G e1 e1' t1
      -> Rv G e2 e2' t2
      -> tvcc Rk Rc Rt Rv G (tm_pair e1 e2) (tm_pair e1' e2') (cn_prod t1 t2)

| tvcc_plam {G k k' t e e'}
    : Rk G k k'
      -> Rv (G; cl_cn k) e e' t
      -> tvcc Rk Rc Rt Rv G (tm_plam k e) (tm_plam k' e') (cn_all k t)

| tvcc_pack {G k k' t t' c c' e e'}
    : Rc G c c' k
      -> Rk G k k'
      -> Rv G e e' (subst1 c t)
      -> Rc (G; cl_cn k) t t' kd_type
      -> tvcc Rk Rc Rt Rv G (tm_pack c e k t) (tm_pack c' e' k' t') (cn_exists k t)

| tvcc_equiv {G t1 t2 e e'}
    : Rv G e e' t1
      -> eqc G t1 t2 kd_type
      -> tvcc Rk Rc Rt Rv G e e' t2
.


Definition tcompatible (Rk : kreln) (Rc : creln) (Rt : treln) (Rv : treln)
  :=
  (forall G e e' t, wf G -> ttcc Rk Rc Rt Rv G e e' t -> Rt G e e' t)
  /\
  (forall G e e' t, wf G -> tvcc Rk Rc Rt Rv G e e' t -> Rv G e e' t)
.


(* Congruences *)

Definition tsregular
  (Rk : kreln)
  (Rc : creln)
  :=
  (forall G k k', wf G -> Rk G k k' -> eqk G k k')
  /\
  (forall G c c' k, wf G -> Rc G c c' k -> eqc G c c' k).


Definition tcregular
  (Rt : treln)
  (Rv : treln)
  :=
  (forall G e e' t, wf G -> Rt G e e' t -> oft G e t /\ oft G e' t /\ target e /\ target e')
  /\
  (forall G e e' t, wf G -> Rv G e e' t -> oft G e t /\ oft G e' t /\ svalue e /\ svalue e' /\ target e /\ target e').


Definition tcreflexive
  (Rt : treln)
  (Rv : treln)
  :=
  (forall G e t, oft G e t -> target e -> Rt G e e t)
  /\
  (forall G e t, oft G e t -> svalue e -> target e -> Rv G e e t).

  
Definition tssymmetric
  (Rk : kreln)
  (Rc : creln)
  :=
  (forall G k k', Rk G k k' -> Rk G k' k)
  /\
  (forall G c c' k, Rc G c c' k -> Rc G c' c k).


Definition tcsymmetric
  (Rt : treln)
  (Rv : treln)
  :=
  (forall G t e e', Rt G e e' t -> Rt G e' e t)
  /\
  (forall G t e e', Rv G e e' t -> Rv G e' e t).


Definition tctransitive
  (Rt : treln)
  (Rv : treln)
  :=
  (forall G t e1 e2 e3, Rt G e1 e2 t -> Rt G e2 e3 t -> Rt G e1 e3 t)
  /\
  (forall G t e1 e2 e3, Rv G e1 e2 t -> Rv G e2 e3 t -> Rv G e1 e3 t).


Definition tssubstitutive
  (Rk : kreln)
  (Rc : creln)
  :=
  (forall G1 s G2 k k',
     subof sof G1 s G2 
     -> wf G2
     -> Rk G2 k k'
     -> Rk G1 (subst s k) (subst s k'))
  /\
  (forall G1 s G2 c c' k,
     subof sof G1 s G2 
     -> wf G2
     -> Rc G2 c c' k
     -> Rc G1 (subst s c) (subst s c') (subst s k)).


Definition tcsubstitutive (Rt : treln) (Rv : treln)
  :=
  (forall G1 s G2 e e' t,
     subof aof G1 s G2 
     -> targetsub s
     -> wf G2
     -> Rt G2 e e' t
     -> Rt G1 (subst s e) (subst s e') (subst s t))
  /\
  (forall G1 s G2 e e' t,
     subof aof G1 s G2 
     -> targetsub s
     -> wf G2
     -> Rv G2 e e' t
     -> Rv G1 (subst s e) (subst s e') (subst s t)).


(* Congruence minus symmetry. *)
Definition tsemicongruence Rt Rv :=
  tcregular Rt Rv
  /\ tcreflexive Rt Rv
  /\ tctransitive Rt Rv
  /\ tcsubstitutive Rt Rv
  /\ tcompatible samek samec Rt Rv.


Definition tcongruence Rt Rv :=
  tsemicongruence Rt Rv
  /\ tcsymmetric Rt Rv.


(* Congruence minus symmetry and transitivity. *)
Definition tquasicongruence Rt Rv :=
  tcregular Rt Rv
  /\ tcreflexive Rt Rv
  /\ tcsubstitutive Rt Rv
  /\ tcompatible samek samec Rt Rv.


Definition tconsistent
  (Rt : context -> (* tm *) term -> (* tm *) term -> (* cn *) term -> Prop)
  :=
  forall e e' t,
    Rt nil e e' t
    -> thalt nil e
    -> thalt nil e'.


(* Properties *)

Lemma same_tsregular :
  tsregular samek samec.
Proof.
split.
  intros G k k' _ H.
  invert H; [].
  intros; subst; eauto using eqk_refl; done.

  intros G c c' k _ H.
  invert H; [].
  intros; subst; eauto using eqc_refl; done.
Qed.


Lemma eq_tsregular :
  tsregular eqk eqc.
Proof.
unfold tsregular; auto.
Qed.


Lemma same_tssubstitutive :
  tssubstitutive samek samec.
Proof.
split.
  eauto using subst_samek; done.

  eauto using subst_samec; done.
Qed.


Lemma eq_tssubstitutive :
  tssubstitutive eqk eqc.
Proof.
split.
  eauto using subst_eqk; done.

  eauto using subst_eqc; done.
Qed.


Lemma same_tssymmetric :
  tssymmetric samek samec.
Proof.
split.
  intros G k k' H.
  invert H; intros; subst; eauto with same; done.

  intros G c c' k H.
  invert H; intros; subst; eauto with same; done.
Qed.  


Lemma eq_ssymmetric :
  tssymmetric eqk eqc.
Proof.
split; eauto with static.
Qed.  


Lemma ttcc_mono :
  forall (Rk Rk' : kreln) (Rc Rc' : creln) (Rt Rt' : treln) (Rv Rv' : treln),
    (forall G k k', Rk G k k' -> Rk' G k k')
    -> (forall G c c' k, Rc G c c' k -> Rc' G c c' k)
    -> (forall G e e' t, Rt G e e' t -> Rt' G e e' t)
    -> (forall G e e' t, Rv G e e' t -> Rv' G e e' t)
    -> forall G e e' t,
         ttcc Rk Rc Rt Rv G e e' t
         -> ttcc Rk' Rc' Rt' Rv' G e e' t.
Proof.
intros Rk Rk' Rc Rc' Rt Rt' Tv Rv' Himpk Himpc Himpt Himpv G e e' t H.
induct H; intros; eauto 6 using ttcc.
Qed.


Lemma tvcc_mono :
  forall (Rk Rk' : kreln) (Rc Rc' : creln) (Rt Rt' : treln) (Rv Rv' : treln),
    (forall G k k', Rk G k k' -> Rk' G k k')
    -> (forall G c c' k, Rc G c c' k -> Rc' G c c' k)
    -> (forall G e e' t, Rt G e e' t -> Rt' G e e' t)
    -> (forall G e e' t, Rv G e e' t -> Rv' G e e' t)
    -> forall G e e' t,
         tvcc Rk Rc Rt Rv G e e' t
         -> tvcc Rk' Rc' Rt' Rv' G e e' t.
Proof.
intros Rk Rk' Rc Rc' Rt Rt' Tv Rv' Himpk Himpc Himpt Himpv G e e' t H.
induct H; intros; eauto 6 using tvcc.
Qed.


Lemma ttcc_mono_wf :
  forall Rk Rc (Rt Rt' : treln) (Rv Rv' : treln),
    tsregular Rk Rc
    -> tcregular Rt Rv
    -> (forall G e e' t, wf G -> Rt G e e' t -> Rt' G e e' t)
    -> (forall G e e' t, wf G -> Rv G e e' t -> Rv' G e e' t)
    -> forall G e e' t,
         wf G
         -> ttcc Rk Rc Rt Rv G e e' t
         -> ttcc Rk Rc Rt' Rv' G e e' t.
Proof.
intros Rk Rc Rt Rt' Rv Rv' Hsreg Hreg Himpt Himpv G e e' t Hwf Hcc.
revert Hwf.
#2 induct Hcc;
  try (intros;
       repeat
         (match goal with
          | H : Rk _ _ _ |- _ =>
              let Hof := fresh in
              let Hof' := fresh
              in
                so (eqk_reg _#3 Hwf (Hsreg andel _#3 Hwf H)) as (Hof & Hof');
                revert H
     
          | H : Rc _ _ _ _ |- _ =>
              let Hof := fresh in
              let Hof' := fresh
              in
                so (eqc_reg _#4 Hwf (Hsreg ander _#4 Hwf H)) as (Hof & Hof' & _);
                revert H
          end);
       intros;
       eauto 6 using ttcc, wf;
       done).

(* unpack *)
intros G k t t' e1 e1' e2 e2' H1 H2 Hoft' Hwf.
so (ofc_ofcr _#3 (oft_reg _#3 Hwf (Hreg andel _#4 Hwf H1 andel))) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
eauto 6 using ttcc, wf; done.

(* lett *)
intros G t1 t2 e1 e1' e2 e2' H1 H2 Hwf.
so (oft_reg _#3 Hwf (Hreg andel _#4 Hwf H1 andel)).
eauto 6 using ttcc, wf; done.
Qed.


Lemma tvcc_mono_wf :
  forall Rk Rc (Rt Rt' : treln) (Rv Rv' : treln),
    tsregular Rk Rc
    -> tcregular Rt Rv
    -> (forall G e e' t, wf G -> Rt G e e' t -> Rt' G e e' t)
    -> (forall G e e' t, wf G -> Rv G e e' t -> Rv' G e e' t)
    -> forall G e e' t,
         wf G
         -> tvcc Rk Rc Rt Rv G e e' t
         -> tvcc Rk Rc Rt' Rv' G e e' t.
Proof.
intros Rk Rc Rt Rt' Rv Rv' Hsreg Hreg Himpt Himpv G e e' t Hwf Hcc.
revert Hwf.
induct Hcc;
  try (intros;
       repeat
         (match goal with
          | H : Rk _ _ _ |- _ =>
              let Hof := fresh in
              let Hof' := fresh
              in
                so (eqk_reg _#3 Hwf (Hsreg andel _#3 Hwf H)) as (Hof & Hof');
                revert H
     
          | H : Rc _ _ _ _ |- _ =>
              let Hof := fresh in
              let Hof' := fresh
              in
                so (eqc_reg _#4 Hwf (Hsreg ander _#4 Hwf H)) as (Hof & Hof' & _);
                revert H
          end);
       intros;
       eauto 6 using tvcc, wf;
       done); done.
Qed.


Lemma tcompatible_mono :
  forall (Rk Rk' : _ -> _ -> _ -> Prop) (Rc Rc' : _ -> _ -> _ -> _ -> Prop) Rt Rv,
    (forall G k k', Rk G k k' -> Rk' G k k')
    -> (forall G c c' k, Rc G c c' k -> Rc' G c c' k)
    -> tcompatible Rk' Rc' Rt Rv
    -> tcompatible Rk Rc Rt Rv.
Proof.
intros Rk Rk' Rc Rc' Rt Rv Himplk Himplc (Htcomp & Hvcomp).
split; eauto using ttcc_mono, tvcc_mono; done.
Qed.


Lemma tcompatible_ext :
  forall Rk Rc Rt Rt' Rv Rv',
    (forall G e e' t, Rt G e e' t <-> Rt' G e e' t)
    -> (forall G e e' t, Rv G e e' t <-> Rv' G e e' t)
    -> tcompatible Rk Rc Rt Rv
    -> tcompatible Rk Rc Rt' Rv'.
Proof.
intros Rk Rc Rt Rt' Rv Rv' Hifft Hiffv Hcomp.
so (fun G e e' t => Hifft G e e' t ander) as Himplt.
so (fun G e e' t => Hiffv G e e' t ander) as Himplv.
split.
  intros G e e' t Hwf Httcc.
  apply Hifft; apply Hcomp; auto; [].
  invert Httcc; intros; subst; eauto using ttcc; done.

  intros G e e' t Hwf Htvcc.
  apply Hiffv; apply Hcomp; auto; [].
  invert Htvcc; intros; subst; eauto using tvcc; done.
Qed.


Lemma tsemicongruence_ext :
  forall Rt Rt' Rv Rv',
    (forall G e e' t, Rt G e e' t <-> Rt' G e e' t)
    -> (forall G e e' t, Rv G e e' t <-> Rv' G e e' t)
    -> tsemicongruence Rt Rv
    -> tsemicongruence Rt' Rv'.
Proof.
intros Rt Rt' Rv Rv' Hifft Hiffv ((Htreg & Hvreg) & (Htrefl & Hvrefl) & (Httrans & Hvtrans) & (Htsubst & Hvsubst) & Hcompat).
so (fun G e e' t => Hifft G e e' t ander) as Himplt.
so (fun G e e' t => Hiffv G e e' t ander) as Himplv.
so (fun G e e' t => Hifft G e e' t andel) as Himplt'.
so (fun G e e' t => Hiffv G e e' t andel) as Himpl'.
do2 4 split.
  split; auto; done.

  split; auto; done.

  split; eauto; done.

  split; eauto; done.

  eapply tcompatible_ext; eauto; done.
Qed.


Lemma subst_ttcc :
  forall Rk Rc Rt Rv,
    tsregular Rk Rc
    -> tcregular Rt Rv
    -> tssubstitutive Rk Rc
    -> tcsubstitutive Rt Rv
    -> forall G1 G2 s e e' t,
         subof aof G1 s G2
         -> targetsub s
         -> wf G2
         -> ttcc Rk Rc Rt Rv G2 e e' t
         -> ttcc Rk Rc Rt Rv G1 (subst s e) (subst s e') (subst s t).
Proof.
intros Rk Rc Rt Rv (Hkreg & Hcreg) (Htreg & Hvreg) (Hksubst & Hcsubst) (Htsubst & Hvsubst).
assert (forall G1 s G2 c c', 
          subof sof G1 s G2
          -> wf G2
          -> Rc G2 c c' kd_type -> Rc G1 (subst s c) (subst s c') kd_type) as Hcsubstt.
  intros G1 s G2 c c' Hs Hwf HR.
  replsub kd_type with (subst s kd_type).
  eapply Hcsubst; eauto; done.
intros G1 G2 s e e' t Hs Htar Hwf Httcc.
revert G1 s Hs Htar Hwf.
#7 induct Httcc;
try (intro G; intros;
     repeat
       (match goal with
        | H : Rt G _ _ _ |- _ =>
          let H' := fresh in so (Htsubst _#6 Hs Htar Hwf H) as H'; clear H; simp_sub in H'

        | H : eqc G _ _ _ |- _ =>
          let H' := fresh in so (subst_eqc _#6 (subof_aof_sof _#3 Hs) H) as H'; clear H; simp_sub in H'
        end);
     simp_sub;
     eauto using ttcc; done).

(* lam *)
intros G t1 t1' t2 e e' H1 H2 G1 s Hs Htar Hwf.
simp_sub.
apply ttcc_lam; simp_sub; eauto using subof_aof_sof; [].
repl (subst (compose s sh1) t2) with (subst (dot (var 0) (compose s sh1)) (shift1 t2)) by (simp_sub; reflexivity).
so (eqc_reg _#4 Hwf (Hcreg _#4 Hwf H1) andel).
eapply Htsubst; eauto using subof_bind_aof_tm, under1_target, wf; done.

(* plam *)
intros G k k' t e e' H1 H2 G1 s Hs Htar Hwf.
simp_sub.
so (eqk_reg _#3 Hwf (Hkreg _#3 Hwf H1) andel).
apply ttcc_plam; simp_sub; eauto using subof_aof_sof, subof_bind_aof_cn, wf, subst_tvalue, subof_aof_oof, under1_target; done.

(* papp *)
intros G k t e e' c c' HR H0 G1 s Hs Htar Hwf.
so (Htsubst _#6 Hs Htar Hwf HR) as HR'; simp_sub in HR'.
simp_sub.
rewrite -> split_dot_compose; [].
eapply ttcc_papp; eauto using subof_aof_sof; done.

(* pack *)
intros G k k' t t' c c' e e' H1 H2 HR H3 G1 s Hs Htar Hwf.
so (Htsubst _#6 Hs Htar Hwf HR) as HR'; simp_sub in HR'.
simp_sub.
so (eqk_reg _#3 Hwf (Hkreg _#3 Hwf H2) andel).
eapply ttcc_pack; simp_sub; eauto using subof_aof_sof, subof_bind_aof_cn, wf; done.

(* unpack *)
intros G k t t' e1 e1' e2 e2' HR1 HR2 H1 G1 s Hs Htar Hwf.
so (ofc_ofcr _#3 (oft_reg _#3 Hwf (Htreg _#4 Hwf HR1 andel))) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
so (Htsubst _#6 Hs Htar Hwf HR1) as HR1'; simp_sub in HR1'.
so (Htsubst _#6 (subof_bind_aof_cn_tm _#5 Hs) (under2_target _ Htar) (wf_tm (wf_cn Hwf Hofk) Hoft) HR2) as HR2'; simp_sub in HR2'.
simp_sub.
eapply ttcc_unpack; simp_sub; eauto using subst_ofc_type, subof_aof_sof; done.

(* fix *)
intros G t t' e e' H HR G1 s Hs Htar Hwf.
simp_sub.
apply ttcc_fix; eauto using subof_aof_sof; [].
so (Htsubst _#6 Hs Htar Hwf HR) as HR'; simp_sub in HR'.
assumption.

(* lett *)
intros G t1 t2 e1 e1' e2 e2' HR1 HR2 G1 s Hs Htar Hwf.
so (oft_reg _#3 Hwf (Htreg _#4 Hwf HR1 andel)) as Hoft.
so (Htsubst _#6 Hs Htar Hwf HR1) as HR1'; simp_sub in HR1'.
so (Htsubst _#6 (subof_bind_aof_tm _#4 Hs) (under1_target _ Htar) (wf_tm Hwf Hoft) HR2) as HR2'; simp_sub in HR2'.
simp_sub.
eapply ttcc_lett; simp_sub; eauto; done.
Qed.


Lemma subst_tvcc :
  forall Rk Rc Rt Rv,
    tsregular Rk Rc
    -> tcregular Rt Rv
    -> tssubstitutive Rk Rc
    -> tcsubstitutive Rt Rv
    -> forall G1 G2 s e e' t,
         subof aof G1 s G2
         -> targetsub s
         -> wf G2
         -> tvcc Rk Rc Rt Rv G2 e e' t
         -> tvcc Rk Rc Rt Rv G1 (subst s e) (subst s e') (subst s t).
Proof.
intros Rk Rc Rt Rv (Hkreg & Hcreg) (Htreg & Hvreg) (Hksubst & Hcsubst) (Htsubst & Hvsubst).
assert (forall G1 s G2 c c', 
          subof sof G1 s G2
          -> wf G2
          -> Rc G2 c c' kd_type -> Rc G1 (subst s c) (subst s c') kd_type) as Hcsubstt.
  intros G1 s G2 c c' Hs Hwf HR.
  replsub kd_type with (subst s kd_type).
  eapply Hcsubst; eauto; done.
intros G1 G2 s e e' t Hs Htar Hwf Htvcc.
revert G1 s Hs Htar Hwf.
#3 induct Htvcc;
try (intro G; intros;
     repeat
       (match goal with
        | H : eqc G _ _ _ |- _ =>
          let H' := fresh in so (subst_eqc _#6 (subof_aof_sof _#3 Hs) H) as H'; clear H; simp_sub in H'
        end);
     simp_sub;
     eauto using tvcc; done).

(* lam *)
intros G t1 t1' t2 e e' H1 H2 G1 s Hs Htar Hwf.
simp_sub.
apply tvcc_lam; simp_sub; eauto using subof_aof_sof; [].
repl (subst (compose s sh1) t2) with (subst (dot (var 0) (compose s sh1)) (shift1 t2)) by (simp_sub; reflexivity).
so (eqc_reg _#4 Hwf (Hcreg _#4 Hwf H1) andel).
eapply Htsubst; eauto using subof_bind_aof_tm, wf, under1_target; done.

(* plam *)
intros G k k' t e e' H1 H2 G1 s Hs Htar Hwf.
simp_sub.
so (eqk_reg _#3 Hwf (Hkreg _#3 Hwf H1) andel).
apply tvcc_plam; simp_sub; eauto using subof_aof_sof, subof_bind_aof_cn, wf, subst_tvalue, subof_aof_oof, under1_target; done.

(* pack *)
intros G k k' t t' c c' e e' H1 H2 HR H3 G1 s Hs Htar Hwf.
so (Hvsubst _#6 Hs Htar Hwf HR) as HR'; simp_sub in HR'.
simp_sub.
so (eqk_reg _#3 Hwf (Hkreg _#3 Hwf H2) andel).
eapply tvcc_pack; simp_sub; eauto using subof_aof_sof, subof_bind_aof_cn, wf; done.
Qed.



(* Weak compatible closure *)

Inductive ttccw
  (Rk : context -> (* kd *) term -> (* kd *) term -> Prop)
  (Rc : context -> (* cn *) term -> (* cn *) term -> (* kd *) term -> Prop)
  (Rt : context -> (* tm *) term -> (* tm *) term -> (* cn *) term -> Prop)
  (Rv : context -> (* tm *) term -> (* tm *) term -> (* cn *) term -> Prop)
  : context -> (* tm *) term -> (* tm *) term -> (* cn *) term -> Prop
  :=
| ttccw_lam {G t1 t1' t2 e e'}
    : Rc G t1 t1' kd_type
      -> Rt (G; cl_tm t1) e e' (shift1 t2)
      -> ttccw Rk Rc Rt Rv G (tm_lam t1 e) (tm_lam t1' e') (cn_arrow t1 t2)

| ttccw_app1 {G t1 t2 e1 e1' e2}
    : Rt G e1 e1' (cn_arrow t1 t2)
       -> oft G e2 t1
       -> target e2
       -> ttccw Rk Rc Rt Rv G (tm_app e1 e2) (tm_app e1' e2) t2

| ttccw_app2 {G t1 t2 e1 e2 e2'}
    : oft G e1 (cn_arrow t1 t2)
      -> target e1
       -> Rt G e2 e2' t1
       -> ttccw Rk Rc Rt Rv G (tm_app e1 e2) (tm_app e1 e2') t2

| ttccw_pair1 {G t1 t2 e1 e1' e2}
    : Rt G e1 e1' t1
      -> oft G e2 t2
      -> target e2
      -> ttccw Rk Rc Rt Rv G (tm_pair e1 e2) (tm_pair e1' e2) (cn_prod t1 t2)

| ttccw_pair2 {G t1 t2 e1 e2 e2'}
    : oft G e1 t1
      -> target e1
      -> Rt G e2 e2' t2
      -> ttccw Rk Rc Rt Rv G (tm_pair e1 e2) (tm_pair e1 e2') (cn_prod t1 t2)

| ttccw_pi1 {G t1 t2 e e'}
    : Rt G e e' (cn_prod t1 t2)
      -> ttccw Rk Rc Rt Rv G (tm_pi1 e) (tm_pi1 e') t1

| ttccw_pi2 {G t1 t2 e e'}
    : Rt G e e' (cn_prod t1 t2)
      -> ttccw Rk Rc Rt Rv G (tm_pi2 e) (tm_pi2 e') t2

| ttccw_plam {G k k' t e e'}
    : Rk G k k'
      -> Rv (G; cl_cn k) e e' t
      -> ttccw Rk Rc Rt Rv G (tm_plam k e) (tm_plam k' e') (cn_all k t)

| ttccw_papp {G k t e e' c c'}
    : Rt G e e' (cn_all k t)
      -> Rc G c c' k
      -> ttccw Rk Rc Rt Rv G (tm_papp e c) (tm_papp e' c') (subst1 c t)

| ttccw_pack {G k k' t t' c c' e e'}
    : Rc G c c' k
      -> Rk G k k'
      -> Rt G e e' (subst1 c t)
      -> Rc (G; cl_cn k) t t' kd_type
      -> ttccw Rk Rc Rt Rv G (tm_pack c e k t) (tm_pack c' e' k' t') (cn_exists k t)

| ttccw_unpack1 {G k t t' e1 e1' e2}
    : Rt G e1 e1' (cn_exists k t)
      -> oft (G; cl_cn k; cl_tm t) e2 (subst (sh 2) t')
      -> target e2
      -> ofc G t' kd_type
      -> ttccw Rk Rc Rt Rv G (tm_unpack e1 e2) (tm_unpack e1' e2) t'

| ttccw_unpack2 {G k t t' e1 e2 e2'}
    : oft G e1 (cn_exists k t)
      -> target e1
      -> Rt (G; cl_cn k; cl_tm t) e2 e2' (subst (sh 2) t')
      -> ofc G t' kd_type
      -> ttccw Rk Rc Rt Rv G (tm_unpack e1 e2) (tm_unpack e1 e2') t'

| ttccw_fix {G t t' e e'}
    : Rc G t t' kd_type
      -> Rt G e e' (cn_arrow (cn_arrow cn_unit t) t)
      -> ttccw Rk Rc Rt Rv G (tm_fix t e) (tm_fix t' e') t

| ttccw_lett1 {G t1 t2 e1 e1' e2}
    : Rt G e1 e1' t1
      -> oft (G; cl_tm t1) e2 (shift1 t2)
      -> target e2
      -> ttccw Rk Rc Rt Rv G (tm_lett e1 e2) (tm_lett e1' e2) t2

| ttccw_lett2 {G t1 t2 e1 e2 e2'}
    : oft G e1 t1
      -> target e1
      -> Rt (G; cl_tm t1) e2 e2' (shift1 t2)
      -> ttccw Rk Rc Rt Rv G (tm_lett e1 e2) (tm_lett e1 e2') t2

| ttccw_equiv {G t1 t2 e e'}
    : Rt G e e' t1
      -> eqc G t1 t2 kd_type
      -> ttccw Rk Rc Rt Rv G e e' t2
.


Inductive tvccw
  (Rk : context -> (* kd *) term -> (* kd *) term -> Prop)
  (Rc : context -> (* cn *) term -> (* cn *) term -> (* kd *) term -> Prop)
  (Rt : context -> (* tm *) term -> (* tm *) term -> (* cn *) term -> Prop)
  (Rv : context -> (* tm *) term -> (* tm *) term -> (* cn *) term -> Prop)
  : context -> (* tm *) term -> (* tm *) term -> (* cn *) term -> Prop
  :=
| tvccw_lam {G t1 t1' t2 e e'}
    : Rc G t1 t1' kd_type
      -> Rt (G; cl_tm t1) e e' (shift1 t2)
      -> tvccw Rk Rc Rt Rv G (tm_lam t1 e) (tm_lam t1' e') (cn_arrow t1 t2)

| tvccw_pair1 {G t1 t2 e1 e1' e2}
    : Rv G e1 e1' t1
      -> oft G e2 t2
      -> svalue e2
      -> target e2
      -> tvccw Rk Rc Rt Rv G (tm_pair e1 e2) (tm_pair e1' e2) (cn_prod t1 t2)

| tvccw_pair2 {G t1 t2 e1 e2 e2'}
    : oft G e1 t1
      -> svalue e1
      -> target e1
      -> Rv G e2 e2' t2
      -> tvccw Rk Rc Rt Rv G (tm_pair e1 e2) (tm_pair e1 e2') (cn_prod t1 t2)

| tvccw_plam {G k k' t e e'}
    : Rk G k k'
      -> Rv (G; cl_cn k) e e' t
      -> tvccw Rk Rc Rt Rv G (tm_plam k e) (tm_plam k' e') (cn_all k t)

| tvccw_pack {G k k' t t' c c' e e'}
    : Rc G c c' k
      -> Rk G k k'
      -> Rv G e e' (subst1 c t)
      -> Rc (G; cl_cn k) t t' kd_type
      -> tvccw Rk Rc Rt Rv G (tm_pack c e k t) (tm_pack c' e' k' t') (cn_exists k t)

| tvccw_equiv {G t1 t2 e e'}
    : Rv G e e' t1
      -> eqc G t1 t2 kd_type
      -> tvccw Rk Rc Rt Rv G e e' t2
.


Definition weakly_tcompatible
  (Rk : context -> (* kd *) term -> (* kd *) term -> Prop)
  (Rc : context -> (* cn *) term -> (* cn *) term -> (* kd *) term -> Prop)
  (Rt : context -> (* tm *) term -> (* tm *) term -> (* cn *) term -> Prop)
  (Rv : context -> (* tm *) term -> (* tm *) term -> (* cn *) term -> Prop)
  :=
  (forall G e e' t, wf G -> ttccw Rk Rc Rt Rv G e e' t -> Rt G e e' t)
  /\
  (forall G e e' t, wf G -> tvccw Rk Rc Rt Rv G e e' t -> Rv G e e' t)
.


Lemma weaken_tcompatible :
  forall Rk Rc Rt Rv,
    tcreflexive Rt Rv
    -> tcompatible Rk Rc Rt Rv
    -> weakly_tcompatible Rk Rc Rt Rv.
Proof.
intros Rk Rc Rt Rv (Hrefl & Hvrefl) (Htcomp & Hvcomp).
split.
  intros G e e' t Hwf Httcc.
  revert Hwf.
  induct Httcc; intros; eauto using ttcc; done.

  intros G e e' t Hwf Htvcc.
  revert Hwf.
  induct Htvcc; eauto using tvcc; done.
Qed.


(* Regularization *)

Definition ttreg (Rt : treln) (G : context) (e e' : (* tm *) term) (t : (* cn *) term)
  :=
  oft G e t
  /\ oft G e' t
  /\ target e
  /\ target e'
  /\ Rt G e e' t.


Definition tvreg (Rv : treln) (G : context) (e e' : (* tm *) term) (t : (* cn *) term)
  :=
  oft G e t
  /\ oft G e' t
  /\ svalue e
  /\ svalue e'
  /\ target e
  /\ target e'
  /\ Rv G e e' t.


Lemma ttreg_intro :
  forall (Rt : treln) G e e' t,
    oft G e t
    -> oft G e' t
    -> target e
    -> target e'
    -> Rt G e e' t
    -> ttreg Rt G e e' t.
Proof.
intros Rt G e1 e2 t Hof1 Hof2 Htar1 Htar2 HR.
do2 4 split; auto.
Qed.


Lemma tvreg_intro :
  forall (Rv : treln) G e e' t,
    oft G e t
    -> oft G e' t
    -> svalue e
    -> svalue e'
    -> target e
    -> target e'
    -> Rv G e e' t
    -> tvreg Rv G e e' t.
Proof.
intros Rt G e1 e2 t Hof1 Hof2 Hval Hval' Htar1 Htar2 HR.
do2 4 split; auto.
Qed.


Lemma ttreg_elim :
  forall Rt G e e' t,
    wf G
    -> ttreg Rt G e e' t
    -> Rt G e e' t.
Proof.
intros Rt G e e' t Hwf HR.
destruct HR as (_ & _ & _ & _ & HR').
assumption.
Qed.


Lemma tvreg_elim :
  forall Rv G e e' t,
    wf G
    -> tvreg Rv G e e' t
    -> Rv G e e' t.
Proof.
intros Rv G e e' t Hwf HR.
destruct HR as (_ & _ & _ & _ & _ & _ & HR').
assumption.
Qed.


Lemma ttreg_reg :
  forall Rt G e e' t,
    wf G
    -> ttreg Rt G e e' t
    -> oft G e t /\ oft G e' t /\ target e /\ target e'.
Proof.
intros Rt G e e' t Hwf Hreg.
so Hreg as (? & ? & ? & ? & _).
auto.
Qed.


Lemma tvreg_reg :
  forall Rv G e e' t,
    wf G
    -> tvreg Rv G e e' t
    -> oft G e t /\ oft G e' t /\ svalue e /\ svalue e' /\ target e /\ target e'.
Proof.
intros Rv G e e' t Hwf Hreg.
so Hreg as (H1 & H2 & H3 & H4 & ? & ? & _).
auto 7.
Qed.


Lemma reg_tcregular :
  forall Rt Rv, tcregular (ttreg Rt) (tvreg Rv).
Proof.
intros Rt Rv.
split.
  intros G e e' t Hwf Hreg.
  eapply ttreg_reg; eauto; done.

  intros G e e' t Hwf Hreg.
  eapply tvreg_reg; eauto; done.
Qed.


Lemma reg_tcreflexive :
  forall Rt Rv,
    tcreflexive Rt Rv
    -> tcreflexive (ttreg Rt) (tvreg Rv).
Proof.
intros Rt Rv (Htrefl & Hvrefl).
split.
  intros G e t Hof.
  do2 2 split; eauto; done.

  intros G e t Hof.
  do2 4 split; eauto; done.
Qed.


Lemma reg_tctransitive :
  forall Rt Rv,
    tctransitive Rt Rv
    -> tctransitive (ttreg Rt) (tvreg Rv).
Proof.
intros Rt Rv (Httrans & Hvtrans ).
split.
  intros G t e1 e2 e3 H12 H23.
  destruct (H12) as (Hof & _ & Htar & _ & Hrel12).
  destruct (H23) as (_ & Hof' & _ & Htar' & Hrel23).
  do2 4 split; eauto; done.

  intros G t e1 e2 e3 H12 H23.
  destruct (H12) as (Hof & _ & Hval & _ & Htar & _ & Hrel12).
  destruct (H23) as (_ & Hof' & _ & Hval' & _ & Htar' & Hrel23).
  do2 4 split; eauto; done.
Qed.


Lemma reg_tcsubstitutive :
  forall Rt Rv,
    tcsubstitutive Rt Rv
    -> tcsubstitutive (ttreg Rt) (tvreg Rv).
Proof.
intros Rt Rv (Htsubst & Hvsubst).
split.
  intros G1 s G2 e e' t Hs Htars Hwf HR.
  destruct (HR) as (Hof & Hof' & Htar & Htar' & Hrel).
  do2 4 split; eauto using subst_oft, subof_aof_dof, subst_target; done.

  intros G1 s G2 e e' t Hs Htars Hwf HR.
  destruct (HR) as (Hof & Hof' & Hval & Hval' & Htar & Htar' & Hrel).
  do2 6 split; eauto using subst_oft, subof_aof_dof, subst_svalue, subst_target; done.
Qed.


Lemma reg_tcompatible :
  forall Rk Rc Rt Rv,
    tsregular Rk Rc
    -> tssubstitutive Rk Rc
    -> tcsubstitutive Rt Rv
    -> tcompatible Rk Rc Rt Rv
    -> tcompatible Rk Rc (ttreg Rt) (tvreg Rv).
Proof.
intros Rk Rc Rt Rv (Hregk & Hregc) Hssubst Hsubst (Htcomp & Hvcomp).
split.

Local Ltac desttreg :=
  repeat
    (match goal with
     | H : ttreg _ ?G _ _ _ |- _ =>
       let Hwf' := fresh
       in
         assert (wf G) as Hwf' by (eauto using wf);
         so (ttreg_reg _#5 Hwf' H) as (? & ? & ? & ?);
         so (ttreg_elim _#5 Hwf' H);
         clear H

     | H : tvreg _ ?G _ _ _ |- _ =>
       let Hwf' := fresh
       in
         assert (wf G) as Hwf' by (eauto using wf);
         so (tvreg_reg _#5 Hwf' H) as (? & ? & ? & ? & ? & ?);
         so (tvreg_elim _#5 Hwf' H);
         clear H
     end).

Local Ltac dottreg Htcomp :=
  desttreg; eapply ttreg_intro; eauto using target with static; apply Htcomp; eauto using ttcc.

Local Ltac dotvreg Hvcomp :=
  desttreg; eapply tvreg_intro; eauto using target with static; apply Hvcomp; eauto using tvcc.

(* TERMS *)
intros G e e' t Hwf Httcc.
revert Hwf.
#7 induct Httcc;
try (intros; dottreg Htcomp; done).

(* lam *)
intros G t1 t1' t2 e e' HR1 HR2 Hwf.
so (Hregc _#4 Hwf HR1) as Heqc.
so (eqc_reg _#4 Hwf Heqc) as (Hof1 & Hof1' & _).
so (ttreg_reg _#5 (wf_tm Hwf Hof1) HR2) as (Hofe & Hofe' & Htar & Htar').
#2 eapply ttreg_intro; eauto using target with static.
  #2 apply (@oft_equiv _ (cn_arrow t1' t2)).
    apply oft_lam; auto; [].
    eapply oft_conv_0t; eauto; done.

    apply eqc_symm; [].
    apply eqc_arrow; auto; [].
    apply eqc_refl; [].
    so (subst_ofc _#5 (unshift_tm _ _) (oft_reg _#3 (wf_tm Hwf Hof1) Hofe)) as H; simp_sub in H.
    assumption.

  eapply Htcomp; eauto; apply ttcc_lam; auto; [].
  eapply ttreg_elim; eauto using wf; done.

(* plam *)
intros G k k' t e e' HR1 HR2 Hwf.
so (Hregk _#3 Hwf HR1) as Heqk.
so (eqk_reg _#3 Hwf Heqk) as (Hof1 & Hof1').
so (tvreg_reg _#5 (wf_cn Hwf Hof1) HR2) as (Hofe & Hofe' & Hval & Hval' & Htar & Htar').
#2 eapply ttreg_intro; eauto using target with static.
  #2 apply (@oft_equiv _ (cn_all k' t)).
    apply oft_plam; auto using subst_svalue; [].
    eapply oft_conv_0c; eauto; done.

    apply eqc_symm; [].
    apply eqc_all; auto; [].
    apply eqc_refl; auto.
    exact (oft_reg _#3 (wf_cn Hwf Hof1) Hofe).

  eapply Htcomp; eauto; apply ttcc_plam; auto; [].
  eapply tvreg_elim; eauto using wf; done.

(* papp *)
intros G k t e e' c c' HR1 HR2 Hwf.
so (Hregc _#4 Hwf HR2) as Heqc.
so (eqc_reg _#4 Hwf Heqc) as (Hofc & Hofc' & _).
so (ttreg_reg _#5 Hwf HR1) as (Hofe & Hofe' & Htar & Htar').
so (ofc_target _#3 Hofc) as Htarc.
so (ofc_target _#3 Hofc') as Htarc'.
#2 eapply ttreg_intro; eauto using target with static.
  #2 apply (@oft_equiv _ (subst1 c' t)).
    eauto with static; done.

    apply eqc_symm; [].
    eapply funct_cn_type; eauto; [].
    so (ofc_ofcr _#3 (oft_reg _#3 Hwf Hofe)) as (kk & Hofcr).
    invert Hofcr; [].
    intros _ Hoft _; clear kk Hofcr.
    assumption.

  apply Htcomp; eauto; eapply ttcc_papp; eauto; [].
  eapply ttreg_elim; eauto; done.

(* pack *)
intros G k k' t t' c c' e e' HRc HRk HRe HRt Hwf.
so (Hregc _#4 Hwf HRc) as Heqc.
so (eqc_reg _#4 Hwf Heqc) as (Hofc & Hofc' & Hofk).
so (Hregc _#4 (wf_cn Hwf Hofk) HRt) as Heqt.
so (eqc_reg _#4 (wf_cn Hwf Hofk) Heqt) as (Hoft & Hoft' & _).
so (ttreg_reg _#5 Hwf HRe) as (Hofe & Hofe' & Htar & Htar').
so (Hregk _#3 Hwf HRk) as Heqk.
so (ofc_target _#3 Hofc) as Htarc.
so (ofc_target _#3 Hofc') as Htarc'.
#2 eapply ttreg_intro; eauto using target with static.
  apply (@oft_equiv _ (cn_exists k' t')); eauto with static; [].
  #2 eapply oft_pack; eauto with static.
    eapply oft_equiv; eauto; [].
    replsub kd_type with (subst1 c kd_type).
    eapply funct_eqc; eauto; done.

    so (subst_ofc _#5 (subof_conv_0 _#3 (eqk_symm Heqk)) Hoft') as H; simp_sub in H.
    assumption.

  apply Htcomp; eauto; apply ttcc_pack; auto; [].
  eapply ttreg_elim; eauto; done.

(* unpack *)
intros G k t t' e1 e1' e2 e2' HR1 HR2 Hoft' Hwf.
so (ofc_ofcr _#3 (oft_reg _#3 Hwf (ttreg_reg _#5 Hwf HR1 andel))) as (l & Hofcr).
invert Hofcr; intros; [].
dottreg Htcomp; done.

(* fix *)
intros G t t' e e' HRt HRe Hwf.
so (Hregc _#4 Hwf HRt) as Heqt.
so (eqc_reg _#4 Hwf Heqt) as (Hoft & Hoft' & _).
so (ttreg_reg _#5 Hwf HRe) as (Hofe & Hofe' & Htar & Htar').
#2 eapply ttreg_intro; eauto using target with static.
  apply (@oft_equiv _ t'); eauto with static; [].
  apply oft_fix; auto; [].
  eapply oft_equiv; eauto with static; done.

  apply Htcomp; eauto; apply ttcc_fix; auto; [].
  eapply ttreg_elim; eauto; done.

(* lett *)
intros G t1 t2 e1 e1' e2 e2' HR1 HR2 Hwf.
so (oft_reg _#3 Hwf (ttreg_reg _#5 Hwf HR1 andel)).
dottreg Htcomp; done.



(* VALUES *)
intros G e e' t Hwf Htvcc.
revert Hwf.
#3 induct Htvcc;
try (intros; dotvreg Hvcomp; done).

(* lam *)
intros G t1 t1' t2 e e' HR1 HR2 Hwf.
so (Hregc _#4 Hwf HR1) as Heqc.
so (eqc_reg _#4 Hwf Heqc) as (Hof1 & Hof1' & _).
so (ttreg_reg _#5 (wf_tm Hwf Hof1) HR2) as (Hofe & Hofe' & Htar & Htar').
#2 eapply tvreg_intro; eauto using target with static.
  #2 apply (@oft_equiv _ (cn_arrow t1' t2)).
    apply oft_lam; auto; [].
    eapply oft_conv_0t; eauto; done.

    apply eqc_symm; [].
    apply eqc_arrow; auto; [].
    apply eqc_refl; [].
    so (subst_ofc _#5 (unshift_tm _ _) (oft_reg _#3 (wf_tm Hwf Hof1) Hofe)) as H; simp_sub in H.
    assumption.

  eapply Hvcomp; eauto; apply tvcc_lam; auto; [].
  eapply ttreg_elim; eauto using wf; done.

(* plam *)
intros G k k' t e e' HR1 HR2 Hwf.
so (Hregk _#3 Hwf HR1) as Heqk.
so (eqk_reg _#3 Hwf Heqk) as (Hof1 & Hof1').
so (tvreg_reg _#5 (wf_cn Hwf Hof1) HR2) as (Hofe & Hofe' & Hval & Hval' & Htar & Htar').
#2 eapply tvreg_intro; eauto using target with static.
  #2 apply (@oft_equiv _ (cn_all k' t)).
    apply oft_plam; auto; [].
    eapply oft_conv_0c; eauto; done.

    apply eqc_symm; [].
    apply eqc_all; auto; [].
    apply eqc_refl; auto.
    exact (oft_reg _#3 (wf_cn Hwf Hof1) Hofe).

  eapply Hvcomp; eauto; apply tvcc_plam; auto; [].
  eapply tvreg_elim; eauto using wf; done.

(* pack *)
intros G k k' t t' c c' e e' HRc HRk HRe HRt Hwf.
so (Hregc _#4 Hwf HRc) as Heqc.
so (eqc_reg _#4 Hwf Heqc) as (Hofc & Hofc' & Hofk).
so (Hregc _#4 (wf_cn Hwf Hofk) HRt) as Heqt.
so (eqc_reg _#4 (wf_cn Hwf Hofk) Heqt) as (Hoft & Hoft' & _).
so (tvreg_reg _#5 Hwf HRe) as (Hofe & Hofe' & Hval & Hval' & Htar & Htar').
so (Hregk _#3 Hwf HRk) as Heqk.
so (ofc_target _#3 Hofc) as Htarc.
so (ofc_target _#3 Hofc') as Htarc'.
#2 eapply tvreg_intro; eauto using target with static.
  apply (@oft_equiv _ (cn_exists k' t')); eauto with static; [].
  #2 eapply oft_pack; eauto with static.
    eapply oft_equiv; eauto; [].
    replsub kd_type with (subst1 c kd_type).
    eapply funct_eqc; eauto; done.

    so (subst_ofc _#5 (subof_conv_0 _#3 (eqk_symm Heqk)) Hoft') as H; simp_sub in H.
    assumption.

  apply Hvcomp; eauto; apply tvcc_pack; auto; [].
  eapply tvreg_elim; eauto; done.
Qed.


Lemma reg_tquasicongruence :
  forall Rt Rv,
    tcreflexive Rt Rv
    -> tcsubstitutive Rt Rv
    -> tcompatible samek samec Rt Rv
    -> tquasicongruence (ttreg Rt) (tvreg Rv).
Proof.
intros Rt Rv Hrefl Hsub Hcomp.
do2 3 split;
auto using reg_tcregular, reg_tcreflexive, reg_tcsubstitutive, reg_tcompatible, same_tsregular, same_tssubstitutive.
Qed.


Lemma reg_tsemicongruence :
  forall Rt Rv,
    tcreflexive Rt Rv
    -> tctransitive Rt Rv
    -> tcsubstitutive Rt Rv
    -> tcompatible samek samec Rt Rv
    -> tsemicongruence (ttreg Rt) (tvreg Rv).
Proof.
intros Rt Rv Hrefl Htrans Hsub Hcomp.
do2 4 split;
auto using reg_tcregular, reg_tcreflexive, reg_tctransitive, reg_tcsubstitutive, reg_tcompatible, same_tsregular, same_tssubstitutive.
Qed.


Lemma reg_tconsistent :
  forall (Rt : _ -> _ -> _ -> _ -> Prop),
    (forall e e' t,
       oft nil e t
       -> oft nil e' t
       -> Rt nil e e' t
       -> thalt nil e
       -> thalt nil e')
    -> tconsistent (ttreg Rt).
Proof.
intros Rt Hcon.
intros e e' t HR Hhalt.
destruct HR as (Hofe & Hofe' & _ & _ & HR).
eauto.
Qed.


Lemma ttcc_formation :
  forall Rt Rv G e e' t,
    tcregular Rt Rv
    -> wf G
    -> ttcc eqk eqc Rt Rv G e e' t
    -> oft G e t /\ oft G e' t /\ target e /\ target e'.
Proof.
intros Rt Rv G e e' t Hcreg Hwf Httcc.
so (reg_tcregular (fun _ _ _ _ => True) (fun _ _ _ _ => True) andel) as Hreg.
eapply Hreg; eauto; [].
#6 apply (reg_tcompatible eqk eqc (fun _ _ _ _ => True) (fun _ _ _ _ => True)).
  exact eq_tsregular.

  exact eq_tssubstitutive.

  split; intros; trivial; done.

  split; intros; trivial; done.

  assumption.

  #3 apply (ttcc_mono_wf eqk eqc Rt _ Rv); auto.
    apply eq_tsregular; done.

    clear G e e' t Hwf Httcc Hreg.
    intros G e e' t Hwf HR.
    so (Hcreg andel _#4 Hwf HR) as (? & ? & ? & ?).
    do2 2 split; auto; done.

    clear G e e' t Hwf Httcc Hreg.
    intros G e e' t Hwf HR.
    so (Hcreg ander _#4 Hwf HR) as (? & ? & ? & ? & ? & ?).
    do2 4 split; auto; done.
Qed.


Lemma tvcc_formation :
  forall Rt Rv G e e' t,
    tcregular Rt Rv
    -> wf G
    -> tvcc eqk eqc Rt Rv G e e' t
    -> oft G e t /\ oft G e' t /\ svalue e /\ svalue e' /\ target e /\ target e'.
Proof.
intros Rt Rv G e e' t Hcreg Hwf Httcc.
so (reg_tcregular (fun _ _ _ _ => True) (fun _ _ _ _ => True) ander) as Hreg.
eapply Hreg; eauto; [].
#6 apply (reg_tcompatible eqk eqc (fun _ _ _ _ => True) (fun _ _ _ _ => True)).
  exact eq_tsregular.

  exact eq_tssubstitutive.

  split; intros; trivial; done.

  split; intros; trivial; done.

  assumption.

  #3 apply (tvcc_mono_wf eqk eqc Rt _ Rv); auto.
    apply eq_tsregular; done.

    clear G e e' t Hwf Httcc Hreg.
    intros G e e' t Hwf HR.
    so (Hcreg andel _#4 Hwf HR) as (? & ? & ? & ?).
    do2 2 split; auto; done.

    clear G e e' t Hwf Httcc Hreg.
    intros G e e' t Hwf HR.
    so (Hcreg ander _#4 Hwf HR) as (? & ? & ? & ? & ? & ?).
    do2 4 split; auto; done.
Qed.


(* Symmetrization *)

Lemma transpose_tsregular :
  forall Rk Rc,
    tsregular Rk Rc
    -> tsregular (fun G k k' => Rk G k' k) (fun G c c' k => Rc G c' c k).
Proof.
intros Rk Rc Hreg.
split.
  intros.
  apply eqk_symm; [].
  eapply (Hreg andel); eauto; done.

  intros.
  apply eqc_symm; [].
  eapply (Hreg ander); eauto; done.
Qed.


Lemma transpose_tssubstitutive :
  forall Rk Rc,
    tssubstitutive Rk Rc
    -> tssubstitutive (fun G k k' => Rk G k' k) (fun G c c' k => Rc G c' c k).
Proof.
intros Rk Rc (Hk & Hc).
split; eauto.
Qed.


Lemma transpose_tcsubstitutive :
  forall Rt Rv,
    tcsubstitutive Rt Rv
    -> tcsubstitutive (fun G e e' t => Rt G e' e t) (fun G e e' t => Rv G e' e t).
Proof.
intros Rt Rv (Htsubst & Hvsubst).
split; eauto; done.
Qed.
    

Lemma transpose_tcompatible :
  forall Rk Rc Rt Rv,
    tsregular Rk Rc
    -> tcregular Rt Rv
    -> tssubstitutive Rk Rc
    -> tcsubstitutive Rt Rv
    -> (forall G c c' k k', eqk G k k' -> Rc G c c' k -> Rc G c c' k')
    -> tssymmetric Rk Rc
    -> tcompatible Rk Rc Rt Rv
    -> tcompatible 
       (fun G k k' => Rk G k' k)
       (fun G c c' k => Rc G c' c k)
       (fun G e e' t => Rt G e' e t)
       (fun G e e' t => Rv G e' e t).

Proof.
intros Rk Rc Rt Rv Hsregular (Htreg & Htvreg) Hssubst Hsubst Hconvert (Hksymm & Hcsymm) (Htcomp & Hvcomp).
split.


(* TERMS *)
cut (forall G e e' t,
       wf G
       -> ttcc
          (fun G k k' => Rk G k' k)
          (fun G c c' k => Rc G c' c k)
          (fun G e e' t => Rt G e' e t)
          (fun G e e' t => Rv G e' e t)
          G e e' t
       -> Rt G e' e t).
  intros H G e e' t Hcc.
  apply H; eauto using subst_ttcc, transpose_tsregular, transpose_tssubstitutive, transpose_tcsubstitutive; done.
destruct Hsregular as (Hkreg & Hcreg).
destruct Hsubst as (Htsubst & Hvsubst).
intros G e e' t Hwf Httcc.
revert Hwf.
#5 induct Httcc;
try (intros; eauto using ttcc; done).

(* lam *)
intros G t1 t1' t2 e e' HRc HR Hwf.
so (Hcreg _#4 Hwf HRc) as Heqc.
so (eqc_reg _#4 Hwf Heqc) as (_ & Hoft1 & _).
#2 apply Htcomp; auto; apply (@ttcc_equiv _#5 (cn_arrow t1' t2)).
  apply Htcomp; auto; [].
  apply ttcc_lam; auto with static; [].
  so (Htsubst _#6 (subof_conv_0t _#3 Heqc) (under_id_target _) (wf_tm Hwf Hoft1) HR) as H; simp_sub in H.
  assumption.

  so (subst_ofc _#5 (unshift_tm _ _) (oft_reg _#3 (wf_tm Hwf Hoft1) (Htreg _#4 (wf_tm Hwf Hoft1) HR andel))) as Hoft2; simp_sub in Hoft2.
  eauto with static; done.

(* plam *)
intros G k k' t e e' HRk HR Hwf.
so (Hkreg _#3 Hwf HRk) as Heqk.
so (eqk_reg _#3 Hwf Heqk) as (_ & Hofk).
#2 apply Htcomp; auto; apply (@ttcc_equiv _#5 (cn_all k' t)).
  apply Htcomp; auto; apply ttcc_plam; eauto with static; [].
  so (Hvsubst _#6 (subof_conv_0_aof _#3 Heqk) (under_id_target _) (wf_cn Hwf Hofk) HR) as H; simp_sub in H.
  assumption.

  so (oft_reg _#3 (wf_cn Hwf Hofk) (Htvreg _#4 (wf_cn Hwf Hofk) HR andel)) as Hoft.
  eauto with static; done.

(* papp *)
intros G k t e e' c c' HR HRc Hwf.
#2 apply Htcomp; auto; apply (@ttcc_equiv _#5 (subst1 c' t)).
  apply Htcomp; auto; eapply ttcc_papp; eauto with static; done.

  so (ofc_ofcr _#3 (oft_reg _#3 Hwf (Htreg _#4 Hwf HR andel))) as (l & H).
  invert H; clear H; [].
  intros _ Hoft _.
  repl kd_type with (subst1 c' kd_type) by (simp_sub; reflexivity).
  eapply funct_cn; eauto with static; done.

(* pack *)
intros G k k' t t' c c' e e' HRc HRk HR HRt Hwf.
so (Hcreg _#4 Hwf HRc) as Heqc.
so (Hkreg _#3 Hwf HRk) as Heqk.
so (eqk_reg _#3 Hwf Heqk) as (_ & Hofk).
#2 apply Htcomp; auto; apply (@ttcc_equiv _#5 (cn_exists k' t')).
  #2 apply Htcomp; auto; eapply ttcc_pack; eauto with static.
    apply Htcomp; auto; apply (@ttcc_equiv _#5 (subst1 c t)); auto; [].
    repl kd_type with (subst1 c kd_type) by (simp_sub; reflexivity).
    eapply funct_eqc; eauto with static; done.

    so (Hssubst ander _#6 (subof_conv_0 _#3 Heqk) (wf_cn Hwf Hofk) HRt) as HRt'; simp_sub in HRt'.
    auto; done.

  so (Hcreg _#4 (wf_cn Hwf Hofk) HRt) as Heqt.
  so (subst_eqc _#6 (subof_conv_0 _#3 Heqk) Heqt) as Heqt'; simp_sub in Heqt'.
  eauto with static; done.

(* fix *)
intros G t t' e e' HRc HR Hwf.
so (Hcreg _#4 Hwf HRc) as Heqc.
apply Htcomp; auto; apply (@ttcc_equiv _#5 t'); auto with static; [].
apply Htcomp; auto; apply ttcc_fix; auto with static; [].
apply Htcomp; auto; eapply ttcc_equiv; eauto with static; done.




(* VALUES *)
cut (forall G e e' t,
       wf G
       -> tvcc
          (fun G k k' => Rk G k' k)
          (fun G c c' k => Rc G c' c k)
          (fun G e e' t => Rt G e' e t)
          (fun G e e' t => Rv G e' e t)
          G e e' t
       -> Rv G e' e t).
  intros H G e e' t Hcc.
  apply H; eauto using subst_ttcc, transpose_tsregular, transpose_tssubstitutive, transpose_tcsubstitutive; done.
destruct Hsregular as (Hkreg & Hcreg).
destruct Hsubst as (Htsubst & Hvsubst).
intros G e e' t Hwf Htvcc.
revert Hwf.
#3 induct Htvcc;
try (intros; eauto using tvcc; done).

(* lam *)
intros G t1 t1' t2 e e' HRc HR Hwf.
so (Hcreg _#4 Hwf HRc) as Heqc.
so (eqc_reg _#4 Hwf Heqc) as (_ & Hoft1 & _).
#2 apply Hvcomp; auto; apply (@tvcc_equiv _#5 (cn_arrow t1' t2)).
  apply Hvcomp; auto; [].
  apply tvcc_lam; auto with static; [].
  so (Htsubst _#6 (subof_conv_0t _#3 Heqc) (under_id_target _) (wf_tm Hwf Hoft1) HR) as H; simp_sub in H.
  assumption.

  so (subst_ofc _#5 (unshift_tm _ _) (oft_reg _#3 (wf_tm Hwf Hoft1) (Htreg _#4 (wf_tm Hwf Hoft1) HR andel))) as Hoft2; simp_sub in Hoft2.
  eauto with static; done.

(* plam *)
intros G k k' t e e' HRk HR Hwf.
so (Hkreg _#3 Hwf HRk) as Heqk.
so (eqk_reg _#3 Hwf Heqk) as (_ & Hofk).
#2 apply Hvcomp; auto; apply (@tvcc_equiv _#5 (cn_all k' t)).
  apply Hvcomp; auto; apply tvcc_plam; eauto with static; [].
  so (Hvsubst _#6 (subof_conv_0_aof _#3 Heqk) (under_id_target _) (wf_cn Hwf Hofk) HR) as H; simp_sub in H.
  assumption.

  so (oft_reg _#3 (wf_cn Hwf Hofk) (Htvreg _#4 (wf_cn Hwf Hofk) HR andel)) as Hoft.
  eauto with static; done.

(* pack *)
intros G k k' t t' c c' e e' HRc HRk HR HRt Hwf.
so (Hcreg _#4 Hwf HRc) as Heqc.
so (Hkreg _#3 Hwf HRk) as Heqk.
so (eqk_reg _#3 Hwf Heqk) as (_ & Hofk).
#2 apply Hvcomp; auto; apply (@tvcc_equiv _#5 (cn_exists k' t')).
  #2 apply Hvcomp; auto; eapply tvcc_pack; eauto with static.
    apply Hvcomp; auto; apply (@tvcc_equiv _#5 (subst1 c t)); auto; [].
    repl kd_type with (subst1 c kd_type) by (simp_sub; reflexivity).
    eapply funct_eqc; eauto with static; done.

    so (Hssubst ander _#6 (subof_conv_0 _#3 Heqk) (wf_cn Hwf Hofk) HRt) as HRt'; simp_sub in HRt'.
    auto; done.

  so (Hcreg _#4 (wf_cn Hwf Hofk) HRt) as Heqt.
  so (subst_eqc _#6 (subof_conv_0 _#3 Heqk) Heqt) as Heqt'; simp_sub in Heqt'.
  eauto with static; done.
Qed.


Lemma transpose_tssymmetric :
  forall Rk Rc,
    tssymmetric Rk Rc
    -> tssymmetric (fun G k k' => Rk G k' k) (fun G c c' k => Rc G c' c k).
Proof.
intros Rk Rc (Hk & Hc).
split; eauto.
Qed.


Lemma transpose_tcompatible' :
  forall Rk Rc Rt Rv,
    tsregular Rk Rc
    -> tcregular Rt Rv
    -> tssubstitutive Rk Rc
    -> tcsubstitutive Rt Rv
    -> (forall G c c' k k', eqk G k k' -> Rc G c c' k -> Rc G c c' k')
    -> tssymmetric Rk Rc
    -> tcompatible Rk Rc Rt Rv
    -> tcompatible Rk Rc
       (fun G e e' t => Rt G e' e t)
       (fun G e e' t => Rv G e' e t).
Proof.
intros Rk Rc Rt Rv Hsreg Hreg Hssubst Hsubst Hequiv Hsymm Hcomp.
apply transpose_tcompatible; eauto using transpose_tsregular, transpose_tssubstitutive, transpose_tssymmetric; [].
#2 eapply tcompatible_mono; eauto.
  intros.
  apply (Hsymm andel); auto; done.

  intros.
  apply (Hsymm ander); auto; done.
Qed.


Lemma transpose_tsemicongruence :
  forall Rt Rv,
    tsemicongruence Rt Rv
    -> tsemicongruence (fun G e e' t => Rt G e' e t) (fun G e e' t => Rv G e' e t).
Proof.
intros Rt Rv (Hreg & (Htrefl & Hvrefl) & (Htrans & Hvtrans) & Hsubst & Hcomp).
do2 4 split.
  destruct Hreg as (Htreg & Htvreg).
  split.
    intros G e e' t Hwf HR.
    so (Htreg _#4 Hwf HR) as (H & H' & ? & ?).
    auto; done.

    intros G e e' t Hwf HR.
    so (Htvreg _#4 Hwf HR) as (H1 & H2 & H3 & H4 & ? & ?).
    auto 7; done.

  split; auto; done.

  split; eauto; done.

  destruct Hsubst as (? & ?).
  split; eauto; done.

  apply transpose_tcompatible'; auto using same_tsregular, same_tssubstitutive, same_tssymmetric; [].
  intros G c c' k k' Heqk Hsame.
  invert Hsame; [].
  intros Hofc <-.
  apply samecrefl; eauto with static; done.
Qed.




(* Conjunction *)

Lemma conjunction_tcompatible :
  forall Rk Rc Rt Rt' Rv Rv',
    tcompatible Rk Rc Rt Rv
    -> tcompatible Rk Rc Rt' Rv'
    -> tcompatible Rk Rc (tintersect Rt Rt') (tintersect Rv Rv').
Proof.
intros Rk Rc Rt Rt' Rv Rv' (Htcomp & Hvcomp) (Htcomp' & Hvcomp').
unfold tintersect.
split.
  intros G e e' t Hwf H.
  revert Hwf.
  induct H;
  intros;
  repeat (match goal with
          | H : _ /\ _ |- _ => destruct H
          end);
  split; eauto using Htcomp, ttcc;
  done.

  intros G e e' t Hwf H.
  revert Hwf.
  induct H;
  intros;
  repeat (match goal with
          | H : _ /\ _ |- _ => destruct H
          end);
  split; eauto using Hvcomp, tvcc;
  done.
Qed.


Lemma conjunction_tsemicongruence :
  forall Rt Rt' Rv Rv',
    tsemicongruence Rt Rv
    -> tsemicongruence Rt' Rv'
    -> tsemicongruence (tintersect Rt Rt') (tintersect Rv Rv').
Proof.
intros Rt Rt' Rv Rv'.
unfold tintersect, mintersect.
intros ((Htreg & Htvreg) & (Htrefl & Hvrefl) & (Htrans & Hvtrans) & (Htsubst & Hvsubst) & Hcomp).
intros ((Htreg' & Htvreg') & (Htrefl' & Hvrefl') & (Htrans' & Hvtrans') & (Htsubst' & Hvsubst') & Hcomp').
do2 4 split; [split .. |];
try (intros;
     repeat (match goal with
             | H : _ /\ _ |- _ => destruct H
             end);
     eauto;
     done); [].
apply conjunction_tcompatible; auto; done.
Qed.


Lemma ttreg_transpose :
  forall R G e e' t,
    ttreg (fun G e e' t => R G e e' t) G e e' t
    -> ttreg (fun G e e' t => R G e' e t) G e' e t.
Proof.
intros R G e e' t H.
so H as (Hof & Hof' & Htar & Htar' & HR).
do2 4 split; auto.
Qed.


Lemma tvreg_transpose :
  forall R G e e' t,
    tvreg (fun G e e' t => R G e e' t) G e e' t
    -> tvreg (fun G e e' t => R G e' e t) G e' e t.
Proof.
intros R G e e' t H.
so H as (Hof & Hof' & Hval & Hval' & Htar & Htar' & HR).
do2 6 split; auto.
Qed.


(* Union *)

Lemma union_tcregular :
  forall Rt Rt' Rv Rv',
    tcregular Rt Rv
    -> tcregular Rt' Rv'
    -> tcregular (tunion Rt Rt') (tunion Rv Rv').
Proof.
intros Rt Rt' Rv Rv' Hreg Hreg'.
split.
  intros G e e' t Hwf H.
  #2 destruct H.
    apply (Hreg andel); auto; done.

    apply (Hreg' andel); auto; done.

  intros G e e' t Hwf H.
  #2 destruct H.
    apply (Hreg ander); auto; done.

    apply (Hreg' ander); auto; done.
Qed.


Lemma union_tcreflexive_1 :
  forall Rt Rt' Rv Rv',
    tcreflexive Rt Rv
    -> tcreflexive (tunion Rt Rt') (tunion Rv Rv').
Proof.
intros Rt Rt' Rv Rv' Hrefl.
split.
  intros G e t H.
  left.
  apply (Hrefl andel); auto; done.

  intros G e t H.
  left.
  apply (Hrefl ander); auto; done.
Qed.


Lemma union_tcsubstitutive :
  forall Rt Rt' Rv Rv',
    tcsubstitutive Rt Rv
    -> tcsubstitutive Rt' Rv'
    -> tcsubstitutive (tunion Rt Rt') (tunion Rv Rv').
Proof.
intros Rt Rt' Rv Rv' Hsubst Hsubst'.
split.
  intros G1 s G2 e e' t Hsubof Htar Hwf H.
  destruct H as [H | H].
    left.
    eapply (Hsubst andel); eauto; done.

    right.
    eapply (Hsubst' andel); eauto; done.

  intros G1 s G2 e e' t Hsubof Htar Hwf H.
  destruct H as [H | H].
    left.
    eapply (Hsubst ander); eauto; done.

    right.
    eapply (Hsubst' ander); eauto; done.
Qed.


Lemma union_tcompatible :
  forall Rk Rc Rt Rt' Rv Rv',
    weakly_tcompatible Rk Rc Rt Rv
    -> weakly_tcompatible Rk Rc Rt' Rv'
    -> weakly_tcompatible Rk Rc (tunion Rt Rt') (tunion Rv Rv').
Proof.
unfold tunion.
intros Rk Rc Rt Rt' Rv Rv' (Htcomp & Hvcomp) (Htcomp' & Hvcomp').
split.
  intros G e e' t Hwf H.
  revert Hwf.
  induct H;
  intros;
  (match goal with
   | H : _ \/ _ |- _ => destruct H; [left | right]
   end);
  eauto using ttccw;
  done.

  intros G e e' t Hwf H.
  revert Hwf.
  induct H;
  intros;
  (match goal with
   | H : _ \/ _ |- _ => destruct H; [left | right]
   end);
  eauto using tvccw;
  done.
Qed.


(* Transitive closure *)

Lemma tc_tcreflexive :
  forall Rt Rv,
    tcreflexive Rt Rv
    -> tcreflexive (ttc Rt) (ttc Rv).
Proof.
intros Rt Rv Hrefl.
split.
  intros G e t Hofe Htar.
  apply plus_one; [].
  apply (Hrefl andel); auto; done.

  intros G e t Hofe Htar Hval.
  apply plus_one; [].
  apply (Hrefl ander); auto; done.
Qed.


Lemma tc_tctransitive :
  forall Rt Rv,
    tctransitive (ttc Rt) (ttc Rv).
Proof.
intros Rt Rv.
split.
  intros G t e1 e2 e3 H12 H23.
  eapply plus_trans; eauto; done.

  intros G t e1 e2 e3 H12 H23.
  eapply plus_trans; eauto; done.
Qed.


Lemma tc_tcsubstitutive :
  forall Rt Rv,
    tcsubstitutive Rt Rv
    -> tcsubstitutive (ttc Rt) (ttc Rv).
Proof.
intros Rt Rv Hsubst.
split.
  intros G1 s G2 e e' t Hs Hwf Httc.
  unfold ttc in Httc; unfold ttc.
  eapply plus_mono_map; eauto; [].
  intros e1 e2 He.
  eapply (Hsubst andel); eauto; done.

  intros G1 s G2 e e' t Hs Hwf Httc.
  unfold ttc in Httc; unfold ttc.
  eapply plus_mono_map; eauto; [].
  intros e1 e2 He.
  eapply (Hsubst ander); eauto; done.
Qed.


Lemma tc_tcregular :
  forall Rt Rv,
    tcregular Rt Rv
    -> tcregular (ttc Rt) (ttc Rv).
Proof.
intros Rt Rv (Htreg & Htvreg).
split.
  intros G e1 e2 t Hwf Hplus.
  so (plus_plusr _#4 Hplus) as (e2' & _ & H2).
  destruct Hplus as (e1' & H1 & _).
  do2 3 split.
    exact (Htreg _#4 Hwf H1 andel).

    exact (Htreg _#4 Hwf H2 anderl).

    exact (Htreg _#4 Hwf H1 anderrl).

    exact (Htreg _#4 Hwf H2 anderrr).

  intros G e1 e2 t Hwf Hplus.
  so (plus_plusr _#4 Hplus) as (e2' & _ & H2).
  destruct Hplus as (e1' & H1 & _).
  do2 5 split.
    exact (Htvreg _#4 Hwf H1 andel).

    exact (Htvreg _#4 Hwf H2 anderl).

    exact (Htvreg _#4 Hwf H1 anderrl).

    exact (Htvreg _#4 Hwf H2 anderrrl).

    exact (Htvreg _#4 Hwf H1 anderrr anderl).

    exact (Htvreg _#4 Hwf H2 anderrr anderr).
Qed.


Lemma ttc_reg :
  forall Rt Rv G e e' t,
    tcregular Rt Rv
    -> wf G
    -> ttc Rt G e e' t
    -> oft G e t /\ oft G e' t /\ target e /\ target e'.
Proof.
intros Rt Rv G e e' t Hreg.
revert G e e' t.
exact (tc_tcregular Rt Rv Hreg andel).
Qed.


Lemma ttc_reg_v :
  forall Rt Rv G e e' t,
    tcregular Rt Rv
    -> wf G
    -> ttc Rv G e e' t
    -> oft G e t /\ oft G e' t /\ svalue e /\ svalue e' /\ target e /\ target e'.
Proof.
intros Rt Rv G e e' t Hreg.
revert G e e' t.
exact (tc_tcregular Rt Rv Hreg ander).
Qed.


Lemma tc_tcomp :
  forall Rt Rv,
    tcregular Rt Rv
    -> tcsubstitutive Rt Rv
    -> weakly_tcompatible samek samec Rt Rv
    -> tcompatible samek samec (ttc Rt) (ttc Rv).
Proof.
intros Rt Rv Hreg Hsubst Hwcomp.
split.


(* TERMS *)
intros G e e' t Hwf Httcc.
unfold ttc.
revert Hwf.
destruct Hwcomp as (Htcomp & _).
induct Httcc.

(* lam *)
intros G t1 t1' t2 e e' Hsame HR Hwf.
invert Hsame; clear Hsame; [].
intros Hoft1 <-.
eapply plus_mono_map; eauto using wf; [].
intros e1 e2 H12.
eauto using ttccw, samec; done.

(* app *)
intros G t1 t2 e1 e1' e2 e2' HR1 HR2 Hwf.
#2 apply (plus_trans _#3 (tm_app e1' e2)).
  eapply (plus_mono_map _#3 (fun z => tm_app z _)); eauto; [].
  intros e e' He.
  simpl in He.
  so (ttc_reg _#6 Hreg Hwf HR2) as (Hof & _ & Htar & _).
  eauto using ttccw; done.

  eapply plus_mono_map; eauto; [].
  intros e e' He.
  simpl in He.
  so (ttc_reg _#6 Hreg Hwf HR1) as (_ & ? & _ & ?).
  eauto using ttccw; done.

(* pair *)
intros G t1 t2 e1 e1' e2 e2' HR1 HR2 Hwf.
#2 apply (plus_trans _#3 (tm_pair e1' e2)).
  eapply (plus_mono_map _#3 (fun z => tm_pair z _)); eauto; [].
  intros e e' He.
  simpl in He.
  so (ttc_reg _#6 Hreg Hwf HR2) as (? & _ & ? & _).
  eauto using ttccw; done.

  eapply plus_mono_map; eauto; [].
  intros e e' He.
  simpl in He.
  so (ttc_reg _#6 Hreg Hwf HR1) as (_ & ? & _ & ?).
  eauto using ttccw; done.

(* pi1 *)
intros G t1 t2 e e' HR Hwf.
eapply plus_mono_map; eauto; [].
intros e1 e2 H12.
simpl in H12.
eauto using ttccw; done.

(* pi2 *)
intros G t1 t2 e e' HR Hwf.
eapply plus_mono_map; eauto; [].
intros e1 e2 H12.
simpl in H12.
eauto using ttccw; done.

(* plam *)
intros G k k' t e e' Hsame HR Hwf.
invert Hsame; clear Hsame; [].
intros Hofk <-.
eapply plus_mono_map; eauto using wf; [].
eauto using ttccw, samek; done.

(* papp *)
intros G k t e e' c c' HR Hsame Hwf.
invert Hsame; clear Hsame; [].
intros Hofc <-.
eapply (plus_mono_map _#3 (fun z => tm_papp z c)); eauto; [].
intros e1 e2 H.
simpl in H.
eauto using ttccw, samec; done.

(* pack *)
intros G k k' t t' c c' e e' Hsamec Hsamek HR1 Hsamet Hwf.
invert Hsamec; invert Hsamek; invert Hsamet; [].
intros Hoft <- Hofk <- Hofc <-.
eapply (plus_mono_map _#3 (fun z => tm_pack _ z _ _)); eauto; [].
intros e1 e2 H.
simpl in H.
eauto using ttccw, samec; done.

(* unpack *)
intros G k t t' e1 e1' e2 e2' HR1 HR2 Hoft' Hwf.
so (ttc_reg _#6 Hreg Hwf HR1) as (_ & Hofe1' & _ & ?).
so (ofc_ofcr _#3 (oft_reg _ _ _ Hwf Hofe1')) as (l & Hofcr).
invert Hofcr; [].
intros Hofk Hoft _.
so (ttc_reg _#6 Hreg (wf_tm (wf_cn Hwf Hofk) Hoft) HR2) as (? & _ & ? & _).
#2 apply (plus_trans _#3 (tm_unpack e1' e2)).
  eapply (plus_mono_map _#3 (fun z => tm_unpack z _)); eauto; [].
  intros e e' He.
  simpl in He.
  eauto using ttccw; done.

  eapply plus_mono_map; eauto using wf; [].
  intros e e' He.
  simpl in He.
  eauto using ttccw; done.

(* fix *)
intros G t t' e e' Hsame HR Hwf.
invert Hsame; clear Hsame; [].
intros Hof <-.
eapply plus_mono_map; eauto using wf; [].
intros e1 e2 H12.
eauto using ttccw, samec; done.

(* lett *)
intros G t t' e1 e1' e2 e2' HR1 HR2 Hwf.
so (ttc_reg _#6 Hreg Hwf HR1) as (_ & Hofe1' & _ & ?).
so (oft_reg _#3 Hwf Hofe1') as Hoft.
so (ttc_reg _#6 Hreg (wf_tm Hwf Hoft) HR2) as (? & _ & ? & _).
#2 apply (plus_trans _#3 (tm_lett e1' e2)).
  eapply (plus_mono_map _#3 (fun z => tm_lett z _)); eauto; [].
  intros e e' He.
  simpl in He.
  eauto using ttccw; done.

  eapply plus_mono_map; eauto using wf; [].
  intros e e' He.
  simpl in He.
  eauto using ttccw; done.

(* equiv *)
intros G t1 t2 e e' HR1 Heqc Hwf.
eapply plus_mono; eauto; [].
intros e1 e2 H.
simpl in H.
eauto using ttccw; done.





(* VALUES *)
intros G e e' t Hwf Htvcc.
unfold ttc.
revert Hwf.
destruct Hwcomp as (_ & Hvcomp).
induct Htvcc.

(* lam *)
intros G t1 t1' t2 e e' Hsame HR Hwf.
invert Hsame; clear Hsame; [].
intros Hoft1 <-.
eapply plus_mono_map; eauto using wf; [].
intros e1 e2 H12.
eauto using tvccw, samec; done.

(* pair *)
intros G t1 t2 e1 e1' e2 e2' HR1 HR2 Hwf.
#2 apply (plus_trans _#3 (tm_pair e1' e2)).
  eapply (plus_mono_map _#3 (fun z => tm_pair z _)); eauto; [].
  intros e e' He.
  simpl in He.
  so (ttc_reg_v _#6 Hreg Hwf HR2) as (? & _ & ? & _ & ? & _).
  eauto using tvccw; done.

  eapply plus_mono_map; eauto; [].
  intros e e' He.
  simpl in He.
  so (ttc_reg_v _#6 Hreg Hwf HR1) as (_ & ? & _ & ? & _ & ?).
  eauto using tvccw; done.

(* plam *)
intros G k k' t e e' Hsame HR Hwf.
invert Hsame; clear Hsame; [].
intros Hofk <-.
eapply plus_mono_map; eauto using wf; [].
eauto using tvccw, samek; done.

(* pack *)
intros G k k' t t' c c' e e' Hsamec Hsamek HR1 Hsamet Hwf.
invert Hsamec; invert Hsamek; invert Hsamet; [].
intros Hoft <- Hofk <- Hofc <-.
eapply (plus_mono_map _#3 (fun z => tm_pack _ z _ _)); eauto; [].
intros e1 e2 H.
simpl in H.
eauto using tvccw, samec; done.

(* equiv *)
intros G t1 t2 e e' HR1 Heqc Hwf.
eapply plus_mono; eauto; [].
intros e1 e2 H.
simpl in H.
eauto using tvccw; done.
Qed.


Lemma tc_tsemicongruence_strong :
  forall Rt Rv,
    tcregular Rt Rv
    -> tcreflexive Rt Rv
    -> tcsubstitutive Rt Rv
    -> weakly_tcompatible samek samec Rt Rv
    -> tsemicongruence (ttc Rt) (ttc Rv).
Proof.
intros Rt Rv Hreg Hrefl Hsubst Hcomp.
do2 4 split; auto using tc_tcregular, tc_tcreflexive, tc_tctransitive, tc_tcsubstitutive; [].
apply tc_tcomp; auto; done.
Qed.    


Lemma tc_tsemicongruence :
  forall Rt Rv,
    tquasicongruence Rt Rv
    -> tsemicongruence (ttc Rt) (ttc Rv).
Proof.
unfold tquasicongruence.
intros Rt Rv H.
destruct_all H.
apply tc_tsemicongruence_strong; auto using weaken_tcompatible; done.
Qed.


Lemma tc_tconsistent :
  forall Rt,
    tconsistent Rt
    -> tconsistent (ttc Rt).
Proof.
intros Rt Hcons.
intros e e' t HR Hhalt.
so (plus_star _#4 HR) as Hstar.
clear HR.
revert Hhalt.
induct Hstar; auto; [].
(* step *)
intros e1 e2 e3 HR _ IH Hhalt.
unfold tconsistent in Hcons.
eauto; done.
Qed.


(* Symmetric closure *)

Lemma stc_tcongruence :
  forall Rt Rv,
    tquasicongruence Rt Rv
    -> tcongruence (tstc Rt) (tstc Rv).
Proof.
intros Rt Rv (Hreg & Hrefl & Hsubst & Hcomp).
set (Rt' G e e' t := Rt G e' e t).
set (Rv' G e e' t := Rv G e' e t).
set (RRt := tunion Rt Rt').
set (RRv := tunion Rv Rv').
assert (tcregular Rt' Rv') as Hreg'.
  split.
    intros G e e' t Hwf HR.
    so (Hreg andel _#4 Hwf HR) as (? & ? & ? & ?).
    auto; done.

    intros G e e' t Hwf HR.
    so (Hreg ander _#4 Hwf HR) as (? & ? & ? & ? & ? & ?).
    auto 7; done.
assert (tcsubstitutive Rt' Rv') as Hsubst'.
  split.
    intros; eapply (Hsubst andel); eauto; done.

    intros; eapply (Hsubst ander); eauto; done.
split.
  do2 4 split.
    apply tc_tcregular; apply union_tcregular; auto; done.

    apply tc_tcreflexive; apply union_tcreflexive_1; auto; done.

    apply tc_tctransitive; done.

    apply tc_tcsubstitutive; apply union_tcsubstitutive; auto; done.

    #3 apply tc_tcomp.
      apply union_tcregular; auto; done.

      apply union_tcsubstitutive; auto; done.

      #2 apply union_tcompatible.
        apply weaken_tcompatible; auto; done.

        apply weaken_tcompatible; auto; [].
        apply transpose_tcompatible'; auto using same_tsregular, same_tssubstitutive, same_tssymmetric; [].
        intros G c c' k k' Heqk Hsame.
        invert Hsame; [].
        intros Hof ->.
        apply samecrefl; eauto with static; done.

  (* symmetry *)
  split.
    intros G t e e' Hstc.
    apply plus_transpose; [].
    apply (plus_mono _ (fun e1 e2 => Rt G e1 e2 t \/ Rt G e2 e1 t)); auto; [].
    intros e1 e2 H.
    unfold transpose.
    destruct H; [right | left]; assumption.

    intros G t e e' Hstc.
    apply plus_transpose; [].
    apply (plus_mono _ (fun e1 e2 => Rv G e1 e2 t \/ Rv G e2 e1 t)); auto; [].
    intros e1 e2 H.
    unfold transpose.
    destruct H; [right | left]; assumption.
Qed.    


Lemma tstc_tconsistent :
  forall Rt,
    tconsistent Rt
    -> tconsistent (fun G e e' t => Rt G e' e t)
    -> tconsistent (tstc Rt).
Proof.
intros Rt Hcons Hcons'.
intros e e' t HR Hhalt.
so (plus_star _#4 HR) as HR'.
revert Hhalt.
induct HR'; auto; [].
intros e1 e2 e3 Hstep _ IH H1.
apply IH; [].
destruct Hstep as [Hstep | Hstep].
  eapply Hcons; eauto; done.

  eapply Hcons'; eauto; done.
Qed.    
